from typing import Callable

import dominate.tags
import dominate.util

from ..objects import attribution


def format_ask(url_handler: Callable, *args: dominate.tags.dom_tag,
               blog_attribution: attribution.BlogAttribution = None):
    """Renders an "ask" in HTML with the given data"""
    if not blog_attribution:
        asker_attribution = dominate.tags.p(
            dominate.tags.strong("Anonymous", cls="asker-name"),
            " asked:",
            cls="asker"
        )

        asker_avatar = dominate.tags.img(
            src=url_handler("https://assets.tumblr.com/images/anonymous_avatar_96.gif"),
            loading="lazy",
            cls="avatar asker-avatar image"
        )

    else:
        asker_attribution = dominate.tags.p(
            dominate.tags.a(
                dominate.tags.strong(blog_attribution.name, cls="asker-name"),
                href=url_handler(f"https://{blog_attribution.name}.tumblr.com/"),
                cls="asker-attribution"
            ),
            " asked:",
            cls="asker"
        )

        if blog_attribution.avatar:
            asker_avatar = dominate.tags.img(
                src=url_handler(blog_attribution.avatar[0].url),
                loading="lazy",
                cls="avatar asker-avatar image"
            )
        else:
            asker_avatar = None

    with dominate.tags.div(cls="ask") as ask:
        with dominate.tags.div(cls="ask-body"):
            with dominate.tags.div(cls="ask-header") as ask_header:
                ask_header.add(asker_attribution)

            with dominate.tags.div(cls="ask-content") as ask_content:
                for children in args:
                    ask_content.add(children)

    if asker_avatar:
        ask.add(asker_avatar)

    return ask





import dominate

from .format import Formatter
from .parse import Parser, LayoutParser
from . import exceptions


def format_npf(contents, layouts=None, *_, url_handler=None, skip_cropped_images=False, pretty_html=False):
    """Formats the given NPF blocks into HTML
    
    Parameters:
        contents: Array of npf blocks to be parsed.
        layouts:  Layout information to arrange the NPF blocks into.

        url_handler:
            A function in which all URLs are passed into. Expects a string in return. 
            By default the internal logic will default to lambda url : url

        skip_cropped_images: 
            Whether or not to include cropped images within the HTML output of this function. 
            By default every image within NPF's media blocks are used via the <img>'s srcset attribute
        
        pretty_html: Whether or not to render human readable html 
    """
    contents = Parser(contents).parse()
    if layouts:
        layouts = LayoutParser(layouts).parse()

    try:
        contains_render_errors = False

        formatted = Formatter(
            contents,
            layouts, 
            url_handler=url_handler,
            skip_cropped_images=skip_cropped_images
        ).format()

    except exceptions.RenderErrorDisclaimerError as e:
        contains_render_errors = True
        formatted = e.rendered_result
        assert formatted is not None
    except Exception as e:
        # TODO
        # Something had gone wrong block
        formatted = dominate.tags.div(cls="post-body")
        contains_render_errors = True

    return contains_render_errors, formatted.render(pretty=pretty_html)



"""Miscellaneous helper functions and objects"""


class CursorIterator:
    """Allows iterating through some iterable one-by-one with an extra peek() method """

    def __init__(self, iterable):
        self.__iterable = iterable
        self.__iterable_iter = iter(iterable)

        self.cursor = 0
        self.current = None
        self._iter_length = len(iterable)

    @property
    def _at_end(self):
        """Checks if we have reached the end of the content list"""
        return self._iter_length <= self.cursor

    def next(self):
        """Go to the next element in the iterator and returns True if successful"""
        try:
            self.current = next(self.__iterable_iter)
            self.cursor += 1
        except StopIteration:
            return False

        return True

    def peek(self):
        """Takes a peek at the next element"""
        if self._at_end:
            return False

        return self.__iterable[self.cursor]



from . import misc
from .. import helpers
from .. import objects


class LayoutParser(helpers.CursorIterator):
    def __init__(self, layouts):
        super().__init__(layouts)

        self._ask_indices = []
        self._max_ask_layout_index = None
        self.layouts = []

    def parse(self):
        """Begins the parsing chain and returns the final list of parsed objects"""
        while self.next():
            if self.current["type"] == "ask":
                # We assume that if we got a layout type of "ask", it'd be the first element in the list.
                indices = [i for i in self.current["blocks"]]

                if attribution := self.current.get("attribution"):
                    attribution = misc.parse_attribution(attribution)

                self.layouts.append(
                    objects.layouts.AskLayout(indices, attribution=attribution),
                )

                self._ask_indices.extend(indices)
            elif self.current["type"] == "rows":
                rows = []
                truncate_after = self.current.get("truncate_after")

                for row in self.current["display"]:
                    indices = [index for index in row["blocks"] if index not in self._ask_indices]
                    if indices:
                        # Which display mode should we use?
                        if mode := row.get("mode"):
                            if mode["type"] == "carousel":
                                mode = objects.layouts.DisplayMode.CAROUSEL
                            else:
                                mode = objects.layouts.DisplayMode.UNSUPPORTED
                        else:
                            mode = objects.layouts.DisplayMode.WEIGHTED

                        rows.append(
                            objects.layouts.RowLayout(
                                ranges=indices,
                                display_mode=mode,
                            )
                        )

                self.layouts.append(
                    objects.layouts.Rows(
                        rows=rows,
                        truncate_after=truncate_after
                    )
                )

        return self.layouts



import enum
from typing import NamedTuple, Optional, Sequence

from .attribution import BlogAttribution


class DisplayMode(enum.Enum):
    WEIGHTED = 0
    CAROUSEL = 1
    UNSUPPORTED = 2


class AskLayout(NamedTuple):
    ranges: list[int]
    attribution: Optional[BlogAttribution] = None


class RowLayout(NamedTuple):
    ranges: Sequence[int]
    display_mode: DisplayMode = DisplayMode.WEIGHTED


class Rows(NamedTuple):
    rows: Sequence[RowLayout]
    truncate_after: Optional[int] = None



BASIC_LAYOUT_CSS = """
.post-body {
  display: flex;
  flex-direction: column;
  max-width: 540px;
  margin: 0;
  font-size: 16px;
  white-space: pre-wrap;
  word-break: break-word;
}

@media (min-width: 540px) {
  .post-body {
    width: 540px;
  }
}

.post-body > * {
  margin: 8px;
}

.post-body .image-block {
  margin: 0;
}

.layout-row {
  display: flex;
  flex-direction: row;
  gap: 4px;
  margin: 5px;
}

.image-container, img {
  width: 100%;
  height: 100%;
  object-fit: cover;
}

img {
  max-width: 100%;
}

/* Text blocks */

.indented {
  margin: 10px;
  border-left: 3px solid rgba(0, 0, 0, 0.07);
  padding-left: 10px;
}

.heading1 {
font-size: 1.6em
}

.heading2 {
font-size: 1.3em
}

.heading1, .heading2 {
  margin: 0;
}

.quote {
  font-family: serif;
  font-size: 26px;
}

.inline-small {
  font-size: 12px;
}

.text-block a {
  color: inherit;
}

/* Link Blocks */

.link-block {
  border: 1px;
  border-style: solid;
  border-color: rgba(0,0,0,0.4);

  width: 100%;
}

.poster-container {
  position: relative;
  display: flex;
  height: 240px;
  text-align: center;
}

.poster-overlay-text {
  display: flex;
  align-items: center;
  justify-content: center;

  position: absolute;
  width: 100%;
  height: 100%;

  background: rgba(0,0,0,0.4);
}

.link-block-title {
  margin: 20px 15px;
  font-size: 1.5em;
}

.poster-overlay-text {
  margin: 0;
  color: #f1f1f1;
}

.link-block-description-container {
  padding: 0 10px 15px 10px;
}

.site-name-author-separator {
  margin: 0px 3px;
}

/* Ask */

.ask {
  display: flex;
  flex-direction: row;
  justify-content: space-between;
  margin-bottom: 10px;
  padding: 5px;
  gap: 10px;
}

.ask-header {
  margin-bottom: 6px;
}

.ask-content * {
  margin-bottom: 10px;
}

.asker-avatar {
  width: 48px;
  height: 48px;
  border-radius: 2px;
}

/* Placeholders */
.unsupported-content-block {
  border: 3px solid rgba(0, 0, 0, 0.07);
  padding: 0em 1em;
}

.unsupported-content-block-message * {
  margin: 10px;
}

.unsupported-content-block h1 {
  font-size: 1.6em
}

"""



import urllib.parse
from typing import Callable

import dominate.tags

from .. import objects


def format_link_attribution(attr: objects.attribution.LinkAttribution, url_handler: Callable):
    return dominate.tags.div(
        dominate.tags.a(
            urllib.parse.urlparse(attr.url).hostname,
            href=url_handler(attr.url),
        ),
        cls="link-attribution"
    )


def format_post_attribution(attr: objects.attribution.PostAttribution, url_handler: Callable):
    return dominate.tags.div(
        dominate.tags.a(
            f"From ",
            dominate.tags.b(
                attr.blog.name
            ),
            href=url_handler(attr.url),
        ),
        cls="post-attribution"
    )

def format_app_attribution(attr: objects.attribution.AppAttribution, url_handler: Callable):
    return dominate.tags.div(
        dominate.tags.a(
            f"View on ",
            dominate.tags.b(
                attr.app_name
            ),
            href=url_handler(attr.url),
        ),
        cls="post-attribution"
    )


# See misc.format_ask()
# -------
# def format_ask_attribution():
#   pass



"""Parses NPF Content blocks into python objects

parser = Parser(content_list)
results = parser.parse()

"""

from . import misc
from .. import helpers
from ..objects import inline, text_block, image, link_block, unsupported


class Parser(helpers.CursorIterator):
    """All-in-one parser to process NPF content types"""

    def __init__(self, content):
        """Initializes the parser with a list of content blocks (json objects) to parse"""
        super().__init__(content)
        self.parsed_result = []

    def _parse_text(self, nest_level=0, in_list_grouping=False):
        """Parses a NPF text content block into a TextBlock

        Takes the selected JSON text content block from `self.current` and parses it into a TextObject.
        Accepts a nest_level argument to handle any nesting.

        Args:
            nest_level: An argument representing how nested the current text block is in relations with other text
        blocks. In the original NPF this is represented as the indent_level. You should probably not be calling this
        method with that argument set as anything other than zero.

        Returns:
            'TextBlock' See documentation for that object for more information
        """

        def create_text_block(text_, subtype_, inline_formatting_, nest_=None):
            """Create a TextBlock object based on the data we have parsed"""
            if not nest_:
                nest_ = None

            return text_block.TextBlock(
                text=text_,
                subtype=subtype_,
                nest=nest_,
                inline_formatting=inline_formatting_,
            )

        text = self.current["text"]
        if subtype := self.current.get("subtype"):
            subtype = getattr(text_block.Subtypes, subtype.upper().replace("-", "_"))

        inline_formats = None
        if inline_formatting := self.current.get("formatting"):
            inline_formats = self._parse_inline_text(inline_formatting)

        # Begins the check to see if we have any children in the next element(s)
        nest_array = []
        while peek := self.peek():
            # Our children can only be TextBlock and ones with a set indent_level attr (which implies that they are
            # related to us)
            #
            # When an indent_level attr is set, we should also probably either be one of the list subtypes or a indented
            # block quote subtype. This check shouldn't be needed, however.
            if peek["type"] != "text" or not (indent_level := (peek.get("indent_level") or peek.get("indentLevel"))):
                break

            # If the next element's indent level is higher than ours (stored as nest_level), they are our children.
            # Thus, we'll store them under us.
            if indent_level > nest_level:
                self.next()
                nested_block = self._parse_text(nest_level=nest_level + 1)
                nest_array.append(nested_block)

            else:
                # If not however, then they are either our siblings,  in the same level as our parent,
                # or an even "higher" (or lower with regard to indent_level) level. Therefore, there is nothing else
                # we need to do here
                break

        # Lists of the same type after to us should be merged into a ListGrouping, provided we aren't already in the
        # process of constructing one at our level.
        if not in_list_grouping and subtype in text_block.ListsSubtype:
            list_grouping = [create_text_block(text, subtype, inline_formats, nest_=nest_array)]
            while peek := self.peek():
                if peek_subtype := peek.get("subtype"):
                    peek_subtype = getattr(text_block.Subtypes, peek_subtype.upper().replace("-", "_"))

                # Make sure we are at the same nest level and that we are the same list type
                indent_level = (peek.get("indent_level") or peek.get("indentLevel"))
                indent_level = indent_level or 0

                if peek_subtype != subtype or indent_level != nest_level:
                    break

                self.next()

                list_grouping.append(self._parse_text(in_list_grouping=True, nest_level=nest_level))

            # If we're at the end then we don't have a peek() to compare to. Thus, we'll compare the
            # currently selected text block and the one that we have in the current context
            if list_grouping:
                return text_block.ListGrouping(type=subtype, group=list_grouping)

        return create_text_block(text, subtype, inline_formats, nest_=nest_array)

    @staticmethod
    def route_inline_format(inline_format, start, end):
        inline_type = getattr(inline.FMTTypes, inline_format["type"].upper())

        match inline_type:
            case (inline.FMTTypes.BOLD | inline.FMTTypes.ITALIC |
                  inline.FMTTypes.STRIKETHROUGH | inline.FMTTypes.SMALL):
                return inline.Standard(
                    start=start,
                    end=end,
                    type=inline_type
                )
            case inline.FMTTypes.LINK:
                return inline.Link(
                    start=start,
                    end=end,
                    type=inline_type,
                    url=inline_format["url"]
                )
            case inline.FMTTypes.MENTION:
                blog = inline_format["blog"]
                return inline.Mention(
                    start=start,
                    end=end,
                    type=inline_type,

                    blog_name=blog["name"],
                    blog_uuid=blog["uuid"],
                    blog_url=blog["url"]
                )
            case inline.FMTTypes.COLOR:
                return inline.Color(
                    start=start,
                    end=end,
                    type=inline_type,

                    hex=inline_format["hex"],
                )

    def _parse_inline_text(self, inline_formatting):
        """Parses the inline formatting of a content block into an array of inline fmt objects"""
        inline_formats = []
        inline_formatting_iter = helpers.CursorIterator(inline_formatting)
        while not inline_formatting_iter._at_end:
            inline_formatting_iter.next()

            inline_format = inline_formatting_iter.current
            start, end = inline_format["start"], inline_format["end"]
            current_parsed_inline_fmt = self.route_inline_format(inline_format, start, end)

            overlapping_formats = []
            while peek := inline_formatting_iter.peek():
                p_start, p_end = peek["start"], peek["end"]

                if start == p_start and end == p_end:
                    overlapping_formats.append(self.route_inline_format(peek, p_start, p_end))
                    inline_formatting_iter.next()

                else:
                    # Tumblr's API should return the list of inline fmts sorted. So if even one doesn't match then
                    # we shouldn't have any overlapping ranges with same start and end
                    break

            if overlapping_formats:
                inline_formats.append(
                    inline.TotalOverlaps(
                        type=[current_parsed_inline_fmt] + overlapping_formats,
                        start=start,
                        end=end
                    )
                )
            else:
                inline_formats.append(current_parsed_inline_fmt)

        return inline_formats

    def _parse_image_block(self):
        """Parses a NPF Image Content Block into a ImageBlock NamedTuple"""
        if not isinstance(self.current["media"], list):
            raw_media_list = [self.current["media"]]
        else:
            raw_media_list = self.current["media"]

        media_list = []
        for img in raw_media_list:
            media_list.append(misc.parse_media_block(img))

        alt_text = self.current.get("alt_text")
        if not alt_text:  # Try Camel Case Variant
            alt_text = self.current.get("altText")

        caption = self.current.get("caption")

        if color_block := self.current.get("colors"):
            colors = [color_hex for color_hex in color_block.values()]
        else:
            colors = None

        if attribution := self.current.get("attribution"):
            attribution = misc.parse_attribution(attribution)

        return image.ImageBlock(
            media=media_list,

            alt_text=alt_text,
            caption=caption,
            colors=colors,

            attribution=attribution
        )

    def _parse_link_block(self):
        """Parses a NPF Link Content Block into a LinkBlock NamedTuple"""
        url = self.current["url"]

        # Optional
        title = self.current.get("title")
        description = self.current.get("description")
        author = self.current.get("author")
        site_name = self.current.get("siteName")
        if not site_name:  # Try snake case
            site_name = self.current.get("site_name")


        display_url = self.current.get("displayUrl")
        if not display_url:  # Try snake case
            display_url = self.current.get("display_url")
        
        if raw_poster := self.current.get("poster"):
            if not isinstance(raw_poster, list):
                raw_media_list = [raw_poster]
            else:
                raw_media_list = raw_poster

            poster_media_object = []
            for poster in raw_poster:
                poster_media_object.append(misc.parse_media_block(poster))
        else:
            poster_media_object = None

        return link_block.LinkBlock(
            url=url,

            title=title,
            description=description,
            author=author,
            site_name=site_name,
            display_url=display_url,
            poster=poster_media_object
        )

    def __parse_block(self):
        """Parses a content block and appends the result to self.parsed_result

        Works by routing specific content types to corresponding parse methods
        """

        match self.current["type"]:
            case "text":
                block = self._parse_text()
                self.parsed_result.append(block)
            case "image":
                block = self._parse_image_block()
                self.parsed_result.append(block)
            case "link":
                block = self._parse_link_block()
                self.parsed_result.append(block)
            case _:
                self.parsed_result.append(unsupported.Unsupported(self.current["type"]))

    def parse(self):
        """Begins the parsing chain and returns the final list of parsed objects"""
        while self.next():
            self.__parse_block()

        return self.parsed_result



import dominate.tags

from . import attribution
from .. import objects


def create_srcset(media_blocks, url_handler):
    """Renders an array of media blocks into a usable srcset data. Contains the content's srcset and also poster's

    Returns:
        A tuple of the main content's srcset and also the poster's if available
    """
    main_srcset = []

    for image in media_blocks:
        main_srcset.append(f"{url_handler(image.url)} {image.width}w")

    return main_srcset


def format_image(image_block, row_length=1, url_handler=None, skip_cropped_images=False):
    """Renders a ImageBlock into HTML"""
    container_attributes = {
        "cls": "image-container"
    }

    # Disabled. We cannot generate a gradient background that takes up the same amount of space as the browser-selected
    # responsive image as we do not know which image the browser is going to pick. Therefore, we shall skip this part.

    # if image_block.colors:
    #     colors = [f"#{color}" for color in image_block.colors]
    #     container_attributes["style"] = f"background: linear-gradient(to left bottom, {', '.join(colors)});"

    container = dominate.tags.div(**container_attributes)

    if not url_handler:
        def url_handler(url):
            return url

    processed_media_blocks = []

    # Fetch the media object with the original dimensions to be used as a src= attribute
    original_media = None
    for media in image_block.media:
        if skip_cropped_images and media.cropped:
            continue

        processed_media_blocks.append(media)

        if media.has_original_dimensions:
            original_media = media

    # If for whatever reason we cannot the media object with the original dimensions
    # we'd just use the one with the highest res (default)

    original_media = original_media or processed_media_blocks[0]

    container.add(
        dominate.tags.img(
            src=url_handler(original_media.url),
            srcset=", ".join(create_srcset(processed_media_blocks, url_handler)),
            cls="image", loading="lazy",
            alt=image_block.alt_text or "image",
            sizes=f"(max-width: 540px) {int(100 / row_length)}vh, {int(540 / row_length)}px"
        )
    )

    # Add attribution HTML
    if attr := image_block.attribution:
        if isinstance(attr, objects.attribution.LinkAttribution):
            container.add(attribution.format_link_attribution(attr, url_handler))
        elif isinstance(attr, objects.attribution.PostAttribution):
            container.add(attribution.format_post_attribution(attr, url_handler))
        elif isinstance(attr, objects.attribution.AppAttribution):
            container.add(attribution.format_app_attribution(attr, url_handler))
        else:
            # TODO Add "Unsupported Attribution HTML"
            raise RuntimeError

    # Similar to the reason above, we won't be able to hide the poster image once the main content loads.
    # if poster_srcset:
    #     container.add(dominate.tags.img(
    #         srcset=", ".join(poster_srcset),
    #         cls="poster",
    #         style="position: absolute; left:0; top:0;"
    #
    #     ))

    return container



"""Objects storing data for inline formatting used in NPF's Text Content Block"""

import enum
from typing import NamedTuple, Union


class FMTTypes(enum.Enum):
    BOLD = 0
    ITALIC = 1
    STRIKETHROUGH = 2
    SMALL = 3
    LINK = 4
    MENTION = 5
    COLOR = 6

    TOTAL_OVERLAP_PACKAGE = 7


class Standard(NamedTuple):
    """A tuple storing data on various inline formatting options"""
    type: FMTTypes
    start: int
    end: int


class Link(NamedTuple):
    """A tuple storing data on formatting an inline link"""
    type: FMTTypes
    start: int
    end: int
    url: str


class Mention(NamedTuple):
    """A tuple storing data on formatting an inline mention of a blog"""
    type: FMTTypes
    start: int
    end: int

    blog_name: str
    blog_url: str
    blog_uuid: str


class Color(NamedTuple):
    """A tuple storing data on formatting colored text"""
    type: FMTTypes
    start: int
    end: int
    hex: str


class TotalOverlaps(NamedTuple):
    """A tuple storing data on formatting operations that overlaps from start to finish

    This allows for easily constructing the nested HTML tags that comes out of this.
    """
    type: list[Union[Standard, Link, Mention, Color]]
    start: int
    end: int


INLINE_FMT_TYPES = Union[Standard, Link, Mention, Color, TotalOverlaps]



from typing import NamedTuple, List, Optional

from .attribution import AttributionTypes
from .media_objects import MediaObject


class ImageBlock(NamedTuple):
    media: List[MediaObject]

    alt_text: Optional[str] = None
    caption: Optional[str] = None
    colors: Optional[List[str]] = None

    attribution: Optional[AttributionTypes] = None



from typing import NamedTuple, Optional


class MediaObject(NamedTuple):
    """NamedTuple representing Tumblr's NPF Media Object"""
    url: str  # URL to the media in question
    width: int  # Default: 540
    height: int  # Default: 405

    type: Optional[str] = None  # MIME type

    original_dimensions_missing: bool = None  # If original dimensions are missing
    cropped: bool = None  # If media object is cropped
    has_original_dimensions: bool = None  # If object has the same dimensions as the original media

    poster: 'MediaObject' = None  # still frames of GIFs for low bandwidth users
    video: 'MediaObject' = None  # Video variants of GIFs (?)



import itertools
import queue
from typing import Sequence, Callable

import dominate.tags
import dominate.util

from .. import objects, helpers


class _OperationsIterator(helpers.CursorIterator):
    """Controlled iterator to iterate through a list of inline formatting operations

    Subclasses the CursorIterator object.
    For more information see parent.

    Extends next() to create two additional attributes:
        next_start: Starting index of next operation
        next_end: Ending index of next operation

    """

    def __init__(self, operations):
        super().__init__(operations)
        self.next_start = None
        self.next_end = None

    def next(self):
        """Extends next() to store next_start and next_end"""
        status = super().next()
        if status:  # Save a bit of performance.
            if peek := self.peek():
                self.next_start = peek.start
                self.next_end = peek.end

        if not status:
            self.next_start = None
            self.next_end = None

        return status


class InlineFormatter(helpers.CursorIterator):
    """Formatter for NPF's TextBlock's inline formatting"""

    def __init__(self, string: str, inline_formats: Sequence[objects.inline.INLINE_FMT_TYPES], url_handler: Callable):
        """Initializes InlineFormatter with some string and the formats to apply to it """
        super().__init__(string)
        self.parent_tag = dominate.tags.span(cls="inline-formatted-content")

        # Sorting just in case although the tumblr API should already return
        # sorted formatting
        #
        # Each inline format is going to be called an "operation"

        # Returns a list of operations sorted by the start (ascending) and end (descending) of each
        operations = sorted(inline_formats, key=lambda format_op: (format_op.start, -format_op.end))
        self._ops = _OperationsIterator(operations)

        self.url_handler = url_handler

        self._accumulator = []

        # Temporarily latches onto operations that needs to be completed
        # in the next context for whatever reasons
        self._priority_operation_queue = queue.Queue()

    @property
    def _reached_new_format_operation(self):
        """Check if we have reached the start of a new format operation"""
        if self.cursor in (self._ops.current.start, self._ops.next_start):
            return True
        else:
            return False

    def next(self):
        """Extends next() to append the selected character to the _accumulator attribute"""
        status = super().next()
        if status:
            self._accumulator.append(self.current)

        return status

    def _perform_operation(self, current_tag, till):
        """Performs a formatting option till the specified ending index

        Arguments:
            current_tag: The tag in which further nested tags and text gets applied/added to
            till: An index at which the perform_operation method should break and stop.

        """

        # If this stops then that means either a new formatting section begun (likely while in the midst of ours,
        # aka an overlapping region) or we've reached the end of the string, or the end of the current formatting
        # section (the till argument).
        while not self._reached_new_format_operation and not self._at_end and (self.cursor < till):

            # Handle overlapping regions that starts at the same location as us but end differently.
            #
            # We don't have to check for peek() also having the same end as those are already condensed into
            # a single operation during the parsing stage.
            if self._ops.current.start == self._ops.next_start:
                # If we (operation or till-wise) still last longer than the next operation
                # then they'll be handled under us.
                while till > self._ops.next_end:
                    self._ops.next()
                    self._route_operations(self._ops.current.end, current_tag)

                    # Obviously we can break once our start differs again or we've reached the end
                    if (self._ops.current.start != self._ops.next_start) or self._ops._at_end:
                        break

            self.next()

        # Have we reached our stopping point? (Doesn't matter whether it's the end of our current op or not)
        # If so we dump into the current tag and let whoever called us handle what comes next or
        # what remains of us if any.
        if (self.cursor >= till) or self._at_end:
            current_tag.add(dominate.util.text("".join(self._accumulator)))
            self._accumulator = []

            # Get next operation if we are completely finished
            if till == self._ops.current.end:
                self._ops.next()

            return

        # It should be a new formatting section if we've got this far.
        assert self._reached_new_format_operation

        # So we're going to push all of our accumulated letters to the current tag
        # to get ready for the overlapping section (The next operation will be starting while we're still in the midst
        # of ours.)
        current_tag.add(dominate.util.text("".join(self._accumulator)))
        self._accumulator = []

        try:
            self.next()

            # If our current operation lasts longer than when this method is supposed to end,
            # and a new formatting operation is starting then we need to make sure that this current operation actually
            # gets remembered and finished once the new formatting section has been taken care of.

            # Without this another formatting section in the midst of the one that's about to start might be
            # reached, and we'll lose access to the current formatting section forever.

            operation_latch = None
            if till < self._ops.current.end:
                operation_latch = self._ops.current

            self._ops.next()

            # Do we go on for longer than the (now) current operation?
            #
            # If so we'll take care of the current operation first, and any subsequent operations that are still
            # overlapping with us.
            if till > self._ops.current.end:
                while till > self._ops.current.end and not self._ops._at_end:
                    self._route_operations(self._ops.current.end, parent_tag=current_tag)

                # As the above function breaks as soon as we reach the end of the operations list
                # there's a few considerations we need to take
                if self._ops._at_end:
                    # First we need to finish up the current operation if we aren't there yet
                    if self.cursor < self._ops.current.end:
                        self._route_operations(self._ops.current.end, parent_tag=current_tag)

                    # Since there's no more new operations to consume, the only operations left to finish is our own.
                    # Therefore, we shall advance up until till (either our stopping point or end of our operation)
                    # and then append the result to current_tag and let whoever called us handle the remaining stuff if
                    # any.
                    #
                    # However, sometimes the ending index of an operation (till) is out of bounds. And if we've already reached
                    # the end of the string then there is nothing left to do.
                    while (self.cursor < till) and not self._at_end:
                        self.next()

                    current_tag.add(dominate.util.text("".join(self._accumulator)))
                    self._accumulator = []
                else:
                    # Finish remaining characters until till
                    self._perform_operation(current_tag=current_tag, till=till)

            else:
                self._route_operations(till, parent_tag=current_tag)
                # Read comment above. Dump in priority queue to get fulfilled as soon as we reach route_operations()
                # again, which should be to fulfill the remainder of the current op.
                if operation_latch and till < operation_latch.end:
                    self._priority_operation_queue.put(operation_latch)

        except StopIteration:
            pass

    def _get_tag_of_operation(self, operation):
        """Maps an inline formatting operation to corresponding HTML tags

        This should probably not be used directly. Instead, use _calculate_operation_tags() which calls this method.
        """
        match operation.type:
            case objects.inline.FMTTypes.BOLD:
                return dominate.tags.b(cls="inline-bold")

            case objects.inline.FMTTypes.ITALIC:
                return dominate.tags.i(cls="inline-italics")

            case objects.inline.FMTTypes.STRIKETHROUGH:
                return dominate.tags.s(cls="inline-strikethrough")

            case objects.inline.FMTTypes.SMALL:
                return dominate.tags.small(cls="inline-small")

            case objects.inline.FMTTypes.LINK:
                return dominate.tags.a(href=self.url_handler(operation.url), cls="inline-link")

            case objects.inline.FMTTypes.MENTION:
                return dominate.tags.a(href=self.url_handler(operation.blog_url), cls="inline-mention")

            case objects.inline.FMTTypes.COLOR:
                return dominate.tags.span(style=f"color: {operation.hex};", cls="inline-color")

    def _calculate_operation_tags(self, operation):
        """Converts a specific operation to corresponding HTML tags

        Returns:
            A tuple (working_tag, root_tag), where the working_tag is the tag in which further nested tags and text
            contents should be appended to, and where root_tag is the tag that should be used when appending the final
            product to a parent.

            For instance:
                 We have a root post tag, stored as "post", and we call this method.
                 working_tag, root_tag = self._calculate_operation_tags(operation)

                 The working tag is the one where we apply text contents and further nested tags to:

                 working_tag.add(dominate.util.text("Some text"))

                 And the root tag is the one that we append to the parent post:
                 post.add(root_tag)

            More often than not the working_tag is exactly the same as the root_tag
        """
        if not isinstance(operation.type, list):
            working_tag = self._get_tag_of_operation(operation)
            root_tag = working_tag
        else:
            nested_child_tags = []

            for ops in operation.type:
                nested_child_tags.append(self._get_tag_of_operation(ops))

            for child_tag_one, child_tag_two in itertools.pairwise(nested_child_tags):
                child_tag_one.add(child_tag_two)

            # Operations will be conducted on the innermost tag
            working_tag = nested_child_tags.pop()

            # The outermost nested tag is obviously the root tag
            root_tag = nested_child_tags[0]

        return working_tag, root_tag

    def _perform_priority_operation(self, till, parent_tag):
        """Fetches and performs a priority operation from the priority queue

        The priority operation is an operation that wasn't able to get completed in the last context due to the
        operations iterator advancing past while also having its current context end before reaching the ending point.

        See _perform_operation() for more details on how this method works.
        """

        # Priority queue handling. If something exists within it is likely that the priority operation wasn't able to
        # be completed in the last context due to the operator iterator advancing, and also having reached the till
        # stopping point.
        #
        # For more information see the innards of self.perform_operation
        operation = self._priority_operation_queue.get()
        operation_tag, attachment_tag = self._calculate_operation_tags(operation)

        if operation.start <= self._ops.current.end and self._ops.current.start <= operation.end:
            # If the priority operation ends after the currently selected operation does then we will finish that
            # first. Otherwise, vise versa.
            if operation.end <= self._ops.current.end:
                # Either run until the till value or the end of the priority operation depending on which ends first
                if till < operation.end:
                    self._route_operations(till, parent_tag=operation_tag)

                    # Latch current operation to be handled the next time we reach self.route_operation
                    # since our current context needs to end as we've reached till, and we still haven't reached
                    # operation.end
                    self._priority_operation_queue.put(operation)
                else:
                    self._route_operations(operation.end, parent_tag=operation_tag)

                # If somehow we've neither hit the end of the priority operation, or even till then we'll
                # need to finish off the remainders
                if self.cursor < operation.end and self.cursor < till:
                    # Of course, we might actually have to just run until the till value if it ends before we do
                    if till < operation.end:
                        self._route_operations(till, parent_tag=operation_tag)
                        # So we also latch onto the priority operation as our current context is ending forcing us
                        # to finish in the next context instead
                        self._priority_operation_queue.put(operation)
                    else:
                        self._route_operations(operation.end, parent_tag=operation_tag)

                elif self.cursor < till == self._ops.current.end:
                    # If we have ended, but we still haven't reached the till value then obviously it goes on for
                    # longer than the priority operation. As such we need to finish that.

                    # We just need to make sure that we are working in the same context as what the till
                    # operation should be. It can be checked by a comparison of (till == self.ops.current.end)

                    parent_tag.add(attachment_tag)

                    return self._route_operations(self._ops.current.end, parent_tag=parent_tag)
            else:
                # Since we (the priority operation) last longer than the currently selected operation, we're going
                # to finish it (and any other operations that remains inside us) until we've reached our breaking
                # point of till
                while till < operation.end and not self._ops._at_end:
                    self._route_operations(self._ops.current.end, parent_tag=operation_tag)

                if self._ops._at_end:
                    if self.cursor < till:
                        self._route_operations(till, parent_tag=operation_tag)

                    # We've already processed until the till, and now we're also at the end of the operations list.
                    # There shouldn't be anything more to do. We will directly process until the end of the current
                    # priority operation
                    while self.cursor < operation.end:
                        self.next()

                    operation_tag.add(dominate.util.text("".join(self._accumulator)))
                    self._accumulator = []

                else:
                    # Finish remaining if applicable
                    if self.cursor < till:
                        self._route_operations(till, parent_tag=operation_tag)

                    # If we've reached the end of our current context and still haven't completed the priority
                    # operation then we'll just do it next time we reach self.route_operations
                    if till < operation.end:
                        self._priority_operation_queue.put(operation)
        else:
            # Not Possible
            raise RuntimeError

        if self._at_end:
            return

        return parent_tag.add(attachment_tag)

    def _route_operations(self, till, parent_tag=None):
        """Delegate formatting to specific tags

        Adds result to the given parent_tag or self.parent_tag when unset
        """
        if self._at_end:
            return

        if not parent_tag:
            parent_tag = self.parent_tag

        if not self._priority_operation_queue.empty():
            return self._perform_priority_operation(till, parent_tag)

        working_tag, attachment_tag = self._calculate_operation_tags(self._ops.current)
        self._perform_operation(current_tag=working_tag, till=till)

        parent_tag.add(attachment_tag)

    def format(self):
        """Returns a formatted string formatted with the given inline operations"""

        # Begin operations iteration
        self._ops.next()

        while not self._at_end:
            if self._reached_new_format_operation:
                # Dump the unformatted text we've been collecting into the parent
                self.parent_tag.add(dominate.util.text(''.join(self._accumulator)))
                self._accumulator = []

                till = self._ops.current.end

                # Consumes the next character, so we don't trigger
                # new_format_section again
                self.next()
                self._route_operations(till)

                # Check for remaining characters to format in the current cursor position
                while not self._at_end and (
                    self._ops.current.start <= self.cursor < (till := self._ops.current.end)
                ):
                    self.next()
                    self._route_operations(till)

            # Handle remaining priority queue
            if not self._priority_operation_queue.empty():
                # Inefficient way to get a till value.
                operation = self._priority_operation_queue.get()
                till = operation.end
                self._priority_operation_queue.put(operation)

                # Will reaccess the priority operator queue inside
                self._route_operations(till)

            self.next()

        # Leftovers
        if self._accumulator:
            self.parent_tag.add(dominate.util.text("".join(self._accumulator)))
            self._accumulator = []

        return self.parent_tag



from ..objects import image, attribution


def parse_media_block(media_block):
    """Parses a NPF media object json into a MediaObject NamedTuple"""
    url = media_block["url"]
    mime = media_block.get("type")

    # Defaults to 540x405
    width = media_block.get("width", 540)
    height = media_block.get("height", 405)

    cropped = media_block.get("cropped")

    original_dimensions_missing = media_block.get("original_dimensions_missing")
    has_original_dimensions = media_block.get("has_original_dimensions")

    # Additional checks for the camel case variants
    if not original_dimensions_missing:
        original_dimensions_missing = media_block.get("originalDimensionsMissing")
    if not has_original_dimensions:
        has_original_dimensions = media_block.get("hasOriginalDimensions")

    poster = None
    video = None

    # For some reason they can sometimes be packed into a list with it being the only element.
    if poster_block := media_block.get("poster"):
        poster_block = poster_block[0] if isinstance(poster_block, list) else poster_block
        poster = parse_media_block(poster_block)
    if video_block := media_block.get("video"):
        video_block = video_block[0] if isinstance(video_block, list) else video_block
        video = parse_media_block(video_block)

    return image.MediaObject(
        url=url,
        type=mime,
        width=width,
        height=height,

        original_dimensions_missing=original_dimensions_missing,
        cropped=cropped,
        has_original_dimensions=has_original_dimensions,

        poster=poster,
        video=video
    )


def parse_attribution(attribution_block):
    """Parses a NPF attribution object into corresponding Attribution NamedTuple objects"""
    type_ = attribution_block["type"]

    match type_:
        case "post":
            blog = attribution_block["blog"]
            return attribution.PostAttribution(
                url=attribution_block["url"],
                post=attribution_block["post"]["id"],
                blog=attribution.BlogAttribution(
                    uuid=blog["uuid"],
                    url=blog.get("url"),
                    name=blog.get("name")
                )
            )

        case "link":
            return attribution.LinkAttribution(
                url=attribution_block["url"],
            )
        case "blog":
            blog = attribution_block["blog"]
            if avatars := blog.get("avatar"):
                avatars = [parse_media_block(avatar) for avatar in avatars]
            else:
                avatars = None

            return attribution.BlogAttribution(
                uuid=blog["uuid"],
                url=attribution_block.get("url"),
                name=blog.get("name"),

                avatar=avatars
            )
        case "app":
            if logo := attribution_block.get("logo"):
                logo = parse_media_block(logo)

            url = attribution_block["url"]
            name = attribution_block.get("app_name")
            display_text = attribution_block.get("display_text")

            # Camel case variant checks
            if not name:
                name = attribution_block.get("appName")
            if not display_text:
                display_text = attribution_block.get("displayText")

            return attribution.AppAttribution(
                url=url,
                app_name=name,
                display_text=display_text,
                logo=logo
            )



class RenderErrorDisclaimerError(Exception):
    """Raised when the rendered result may contain errors.

    This is primarily used as a warning. As such the rendered result (wrong as it may be) is
    passed along and can be retrieved.

    """
    def __init__(self, message, rendered_result):
        super().__init__(message)

        self.rendered_result = rendered_result


"""An HTML renderer for the TextBlock object

formatter = TextFormatter(text_block)
formatter.parse()

"""

from typing import Callable

import dominate.tags
import dominate.util

from . import inline
from .. import objects


class TextFormatter:
    """An HTML renderer for the TextBlock object"""

    def __init__(self, text_block: objects.text_block.TextBlock, url_handler: Callable,
                 layout=None, create_list_element=True):
        """Renders a TextBlock object into HTML

        You should probably not be using this class directly!

        Parameters:
            text_block:
                The TextBlock object in question.
            url_handler:
                A callable in which all URLs processed (during inline-formatting) are sent into and retrieved.
                Allows for the caller to alter URLs in the final formatted HTML.
            create_list_element:
                Boolean denoting whether list elements should be enclosed in a <ul> or <ol> element.
                Boolean denoting whether list elements should be enclosed in a <ul> or <ol> element.
                Level zero list merging is delegated to the outer formatter but everything below is handled through
                recursion within the logic of self.format(). See code for more information

        """
        self.text_block = text_block
        self.create_list_element = create_list_element
        self.additional_classes = ""

        if self.text_block.inline_formatting:
            formatter = inline.InlineFormatter(self.text_block.text, self.text_block.inline_formatting, url_handler)
            self.text_tag = formatter.format()

            self.additional_classes += " inline-formatted-block"
        else:
            self.text_tag = dominate.util.text(self.text_block.text)

        self.tag = self.create_tag()

        self.url_handler = url_handler

    def create_tag(self):
        """Create an HTML tag to use as the base for the current TextBlock

        Primary based on the subtype. If nonexistent, then a standard paragraph element is used.

        All elements have the text-block class
        All subtypes have classes corresponding to their types.

        If the text block has inline formatting then an additional inline-formatted class is added
        """
        if not self.text_block.subtype:
            return dominate.tags.p(cls="text-block")

        match self.text_block.subtype:
            case objects.text_block.Subtypes.HEADING1:
                return dominate.tags.h1(cls="text-block heading1" + self.additional_classes)
            case objects.text_block.Subtypes.HEADING2:
                return dominate.tags.h2(cls="text-block heading2" + self.additional_classes)
            case objects.text_block.Subtypes.QUIRKY:
                return dominate.tags.p(cls="text-block quirky" + self.additional_classes)
            case objects.text_block.Subtypes.QUOTE:
                return dominate.tags.blockquote(cls="text-block quote" + self.additional_classes)
            case objects.text_block.Subtypes.INDENTED:
                return dominate.tags.blockquote(cls="text-block indented" + self.additional_classes)
            case objects.text_block.Subtypes.CHAT:
                return dominate.tags.p(cls="text-block chat" + self.additional_classes)
            case objects.text_block.Subtypes.ORDERED_LIST_ITEM:
                return dominate.tags.li(cls="text-block ordered-list-item" + self.additional_classes)
            case objects.text_block.Subtypes.UNORDERED_LIST_ITEM:
                return dominate.tags.li(cls="text-block unordered-list-item" + self.additional_classes)

    def format(self):

        # We have children that needs to be added to us, and we cannot just make the working tag the
        # text contents and add to that. (Can't add to raw text and even if there is an inline-formatting div it still
        # wouldn't be valid HTML in some cases) It needs to be directly added to self.tag. So:
        if self.text_block.nest:
            working_tag = self.tag
            working_tag.add(self.text_tag)
        else:
            working_tag = self.text_tag
            self.tag.add(working_tag)

        # Has nested elements. Also means our subtype is either indented or one of the list types
        if self.text_block.nest:
            for element in self.text_block.nest:
                if isinstance(element, objects.text_block.ListGrouping):
                    if element.type == objects.text_block.Subtypes.ORDERED_LIST_ITEM:
                        list_tag = dominate.tags.ol(cls="ordered-list")
                    else:
                        list_tag = dominate.tags.ul(cls="unordered-list")

                    for block in element.group:
                        list_tag.add(TextFormatter(block, url_handler=self.url_handler).format())

                    working_tag.add(list_tag)
                else:
                    working_tag.add(TextFormatter(element, url_handler=self.url_handler).format())

        return self.tag



from typing import NamedTuple, List, Optional, Union

from .media_objects import MediaObject


class BlogAttribution(NamedTuple):
    # Only UUID is required. The rest are optional but should be present in most cases.
    uuid: str

    url: Optional[str] = None
    name: Optional[str] = None

    avatar: Optional[List[MediaObject]] = None


class PostAttribution(NamedTuple):
    url: str
    post: str
    blog: BlogAttribution


class LinkAttribution(NamedTuple):
    url: str


class AppAttribution(NamedTuple):
    url: str

    app_name: Optional[str] = None
    display_text: Optional[str] = None
    logo: Optional[MediaObject] = None


AttributionTypes = Union[BlogAttribution, PostAttribution, LinkAttribution, AppAttribution]



"""Object representing NPF's Text Content Block

For inline formatting stuff see the adjacent inline.py
"""

import enum
from typing import NamedTuple, Sequence, Union

from .inline import INLINE_FMT_TYPES


class Subtypes(enum.Enum):
    """Enum representing NPF's Text Block's subtype values"""
    HEADING1 = 0
    HEADING2 = 1
    QUIRKY = 2
    QUOTE = 3
    INDENTED = 4
    CHAT = 5
    ORDERED_LIST_ITEM = 6
    UNORDERED_LIST_ITEM = 7


ListsSubtype = (Subtypes.ORDERED_LIST_ITEM, Subtypes.UNORDERED_LIST_ITEM)


class TextBlock(NamedTuple):
    """Object representing Tumblr's NPF's Text Block"""
    text: str
    subtype: Subtypes = None
    nest: 'Sequence[Union[TextBlock, ListGrouping]]' = None

    inline_formatting: Sequence[INLINE_FMT_TYPES] = None


class ListGrouping(NamedTuple):
    """Groups list item TextBlocks of the same type together to create either an ul or ol element"""

    # We reuse ORDERED_LIST_ITEM and UNORDERED_LIST_ITEM to denote which type we should use
    type: Subtypes
    group: Sequence[TextBlock]



from typing import NamedTuple, Optional

from .media_objects import MediaObject


class LinkBlock(NamedTuple):
    url: str

    title: Optional[str] = None
    description: Optional[str] = None
    author: Optional[str] = None
    site_name: Optional[str] = None
    display_url: Optional[str] = None
    poster: Optional[MediaObject] = None



import dominate.tags

from . import text, image, misc
from .. import objects, helpers, exceptions


def _calculate_amount_to_pad_from_nested(block, parent=True):
    if container := getattr(block, "nest", None):
        amount_to_pad = len(block.nest)
    elif container := getattr(block, "group", None):
        # See commit 2e8953d0081ffbe777579fab95a2b98c99d01177 as to why we subtract one
        amount_to_pad = (len(block.group) - 1)
    else:
        return 0

    for child_block in container:
        if getattr(child_block, "nest", None) or getattr(child_block, "group", None):
            amount_to_pad += _calculate_amount_to_pad_from_nested(child_block, parent=False)

    return amount_to_pad


class Formatter(helpers.CursorIterator):
    def __init__(self, content, layout=None, *, url_handler=None, skip_cropped_images = False):
        """Initializes the parser with a list of content blocks (json objects) to parse"""
        super().__init__(content)

        if not url_handler:
            def url_handler(url):
                return url

        self.layout = layout
        self.current_context_padding = 0
        self.render_instructions = []

        self.url_handler = url_handler
        self.skip_cropped_images = skip_cropped_images

        self.has_render_error = False

        self.post = dominate.tags.div(cls="post-body")

    def _format_text(self, block):
        """Formats TextBlock(s) into usable HTML code"""
        formatted_block = text.TextFormatter(block, url_handler=self.url_handler).format()
        return formatted_block

    def _format_list(self, block):
        """Formats a ListGrouping of TextBlock(s) into an HTML list of its corresponding type"""
        if block.type == objects.text_block.Subtypes.ORDERED_LIST_ITEM:
            list_tag = dominate.tags.ol(cls="ordered-list")
        else:
            list_tag = dominate.tags.ul(cls="unordered-list")

        for blk in block.group:
            list_tag.add(self._format_text(blk))

        return list_tag

    def format_unsupported(self, block):
        """Formats a placeholder for unsupported NPF types"""
        self.has_render_error = True

        with dominate.tags.div(cls="unsupported-content-block") as unsupported:
            with dominate.tags.div(cls="unsupported-content-block-message"):
                dominate.tags.h1("Unsupported content placeholder")
                dominate.tags.p(f"Hello! I'm a placeholder for the unsupported \"{block.type}\" type NPF content block."
                                f" Please report me!")

        return unsupported

    def _format_image(self, block, row_length=1):
        """Renders an ImageBlock into HTML"""
        figure = dominate.tags.figure(cls="image-block")
        figure.add(image.format_image(block, row_length, url_handler=self.url_handler, skip_cropped_images=self.skip_cropped_images))

        if block.caption:
            figure.add(dominate.tags.figcaption(block.caption, cls="image-caption"))

        return figure

    def _format_link(self, block):
        container = dominate.tags.div(cls="link-block")
        anchor_content_wrapper = dominate.tags.a(href=self.url_handler(block.url), cls="link-block-link")
        container.add(anchor_content_wrapper)

        if block.poster:
            poster_container = dominate.tags.div(cls="poster-container")
            anchor_content_wrapper.add(poster_container)

            srcset = ", ".join(image.create_srcset(block.poster, self.url_handler))

            poster_container.add(
                dominate.tags.img(
                    srcset=srcset,
                    alt=block.site_name or "Link block poster",
                    sizes="(max-width: 540px) 100vh, 540px",
                )
            )

            if block.title:
                poster_container.add(
                    dominate.tags.div(
                        dominate.tags.span(dominate.util.text(block.title)), cls="link-block-title poster-overlay-text"
                    )
                )
        elif block.title:
            anchor_content_wrapper.add(
                dominate.tags.div(
                    dominate.tags.span(dominate.util.text(block.title)), cls="link-block-title"
                )
            )

        link_block_description_container = dominate.tags.div(cls="link-block-description-container")
        anchor_content_wrapper.add(link_block_description_container)

        if block.description:
            link_block_description_container.add(
                dominate.tags.p(dominate.util.text(block.description), cls="link-block-description")
            )

        if block.site_name or block.author:
            subtitles_div = dominate.tags.div(cls="link-block-subtitles")
            link_block_description_container.add(subtitles_div)

            subtitles = dominate.tags.span()
            subtitles_div.add(subtitles)

            if block.site_name:
                subtitles.add(dominate.tags.span(dominate.util.text(block.site_name)))

                # When an author name exists in addition to the site name then we add it both to the subtitles
                # separated by a |
                if block.author:
                    subtitles.add(dominate.tags.span(dominate.util.text("|"), cls="site-name-author-separator"))
                    subtitles.add(dominate.tags.span(dominate.util.text(block.author)))
            else:
                subtitles.add(dominate.tags.span(dominate.util.text(block.author)))

        return container

    def __prepare_instruction_for_current_block(self):
        """Finds and returns the instruction (method) necessary to render a content block"""
        match self.current:
            case objects.text_block.TextBlock():
                if self.current.nest:
                    self.current_context_padding = _calculate_amount_to_pad_from_nested(self.current)
                return self._format_text, (self.current,)
            case objects.text_block.ListGrouping():
                self.current_context_padding = _calculate_amount_to_pad_from_nested(self.current)

                return self._format_list, (self.current,)
            case objects.image.ImageBlock():
                return self._format_image, (self.current,)
            case objects.link_block.LinkBlock():
                return self._format_link, (self.current,)
            case objects.unsupported.Unsupported():
                return self.format_unsupported, (self.current,)
            case _:  # Unreachable
                raise RuntimeError

    def _pad(self):
        [self.render_instructions.append(None) for _ in range(self.current_context_padding)]
        self.current_context_padding = 0

    def format(self):
        """Renders the list of content blocks into HTML"""
        while self.next():
            instruction = self.__prepare_instruction_for_current_block()
            self.render_instructions.append(instruction)
            self._pad()

        if self.layout:
            # List to keep track of blocks that have been added to layouts
            # this is used to handle the edge case in which only certain blocks have specified layouts but not others.
            blocks_in_layouts = []

            for layout in self.layout:
                if isinstance(layout, objects.layouts.Rows):
                    for row in layout.rows:
                        row_items = []
                        for index in row.ranges:
                            blocks_in_layouts.append(index)

                            render_instructions = self.render_instructions[index]
                            if not render_instructions:
                                continue

                            render_method, arguments = render_instructions

                            match render_method:
                                case self._format_image:
                                    row_items.append(render_method(*arguments, len(row.ranges)))
                                case _:
                                    row_items.append(render_method(*arguments))

                        if not row_items:
                            continue

                        row_tag = dominate.tags.div(cls="layout-row")
                        self.post.add(row_tag)

                        [row_tag.add(i) for i in row_items]
                elif isinstance(layout, objects.layouts.AskLayout):
                    layout_items = []
                    for index in layout.ranges:
                        blocks_in_layouts.append(index)

                        render_instructions = self.render_instructions[index]
                        if not render_instructions:
                            continue

                        render_method, arguments = render_instructions
                        layout_items.append(render_method(*arguments))

                    self.post.add(
                        dominate.tags.div(
                            misc.format_ask(self.url_handler, *layout_items, blog_attribution=layout.attribution),
                            cls="layout-ask")
                    )

                else:
                    continue  # TODO: Unsupported layout type error

            # Edge case:
            # Sometimes only an "ask" layout is specified. In those circumstances the only thing we'll have added to our
            # HTML is the content blocks that makes up the ask. So we'll have some special handling here to handle
            # the leftovers that comes immediately after the ask.
            if len(self.layout) == 1 and isinstance(self.layout[0], objects.layouts.AskLayout):
                for index, render_instructions in enumerate(self.render_instructions):
                    if (index in blocks_in_layouts) or (not render_instructions):
                        continue

                    func, args = render_instructions
                    self.post.add(dominate.tags.div(func(*args), cls="layout-row"))
        else:
            for render_instructions in self.render_instructions:
                if not render_instructions:
                    continue

                func, args = render_instructions
                self.post.add(func(*args))

        if self.has_render_error:
            raise exceptions.RenderErrorDisclaimerError("Rendered post contains errors", rendered_result=self.post)

        return self.post

